# import pytest
#
# from pydriller.repository_mining import RepositoryMining
# from pydriller.tests.integration.concurrency_visitor_test import ConcurrencyVisitorTest
#
# results = []
#
#
# def test_sequential():
#     mv = ConcurrencyVisitorTest()
#     RepositoryMining('test-repos/git-3/', mv).mine()
#     add_result(mv.visited_hashes, mv.visited_times, mv.visited_commits)
#
#
# def test_threads1():
#     mv = ConcurrencyVisitorTest()
#     RepositoryMining('test-repos/git-3/', mv, num_threads=2).mine()
#     add_result(mv.visited_hashes, mv.visited_times, mv.visited_commits)
#
#
# def test_threads2():
#     mv = ConcurrencyVisitorTest()
#     RepositoryMining('test-repos/git-3/', mv, num_threads=100).mine()
#     add_result(mv.visited_hashes, mv.visited_times, mv.visited_commits)
#
#
# def add_result(visited_hashes, visited_times, visited_commits):
#     res = dict()
#     res['visited_hashes'] = visited_hashes
#     res['visited_times'] = visited_times
#     res['visited_commits'] = visited_commits
#     results.append(res)
#
#
# def test_teardown():
#     res1 = results.pop()
#     for result in results:
#         assert res1['visited_hashes'] == result['visited_hashes']
#         assert res1['visited_times'] == result['visited_times']
#         assert res1['visited_commits'] == result['visited_commits']



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import psutil
if 'TRAVIS' in os.environ:
    import logging
    logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s', level=logging.INFO)
from pydriller.repository_mining import RepositoryMining
from datetime import datetime, timezone


def test_memory():
    if 'TRAVIS' not in os.environ:
        return

    p = psutil.Process(os.getpid())
    number_of_commits = 0
    all_commits = []

    dt1 = datetime(2015, 1, 1)
    dt2 = datetime(2016, 1, 1)

    start = datetime.now()
    for _ in RepositoryMining('test-repos/hadoop',
                              since=dt1,
                              to=dt2).traverse_commits():
        memory = p.memory_info()[0] / (2 ** 20)
        all_commits.append(memory)
        number_of_commits += 1

    end = datetime.now()

    diff = end - start
    logging.info('Max memory {} Mb'.format(max(all_commits)))
    logging.info('Min memory {} Mb'.format(min(all_commits)))
    logging.info('All: {}'.format(', '.join(map(str, all_commits))))
    logging.info('Time {}:{}:{}'.format(diff.seconds//3600, (diff.seconds % 3600) // 60, diff.seconds % 60))
    logging.info('Commits per second: {}'.format(len(all_commits) / diff.seconds))


# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from _datetime import datetime
from typing import List
from pydriller.domain.developer import Developer
from pydriller.domain.modification import Modification, ModificationType


class Commit:
    def __init__(self, hash: str, author: Developer, committer: Developer,
                 author_date: datetime, committer_date: datetime,
                 author_timezone: int, committer_timezone: int,
                 msg: str, parents: List[str], merge: bool = False, branches: set = set(),
                 is_commit_in_main_branch: bool = False) -> None:
        """
        Create a commit object.

        :param str hash: hash of the commit
        :param Developer author: author of the commit
        :param Developer committer: committer of the commit
        :param datetime author_date: date when the author committed
        :param datetime committer_date: date when the committer committed
        :param int author_timezone: seconds west from UTC
        :param int committer_timezone: seconds west from UTC
        :param str msg: message of the commit
        :param List[str] parents: list of hashes of the parent commits
        :param bool merge: True if the commit is a merge commit
        :param set branches: branches that include the commit
        :param bool is_commit_in_main_branch: True if the commit is in the main branch
        """
        self.hash = hash
        self.author = author
        self.committer = committer
        self.author_date = author_date
        self.committer_date = committer_date
        self.author_timezone = author_timezone
        self.committer_timezone = committer_timezone
        self.msg = msg
        self.parents = parents
        self.merge = merge
        self.modifications = []  # type: List[Modification]
        self.branches = branches
        self.in_main_branch = is_commit_in_main_branch

    def add_modifications(self, old_path: str, new_path: str, change: ModificationType, diff: str, sc: str):
        """
        Add a modification to the commit.

        :param str old_path: old path of the file (can be null if the file is added)
        :param str new_path: new path of the file (can be null if the file is deleted)
        :param ModificationType change: type of the change
        :param str diff: diff of the change
        :param str sc: source code of the file (can be null if the file is deleted)
        """
        m = Modification(old_path, new_path, change, diff, sc)
        self.modifications.append(m)

    def __eq__(self, other):
        if not isinstance(other, Commit):
            return NotImplemented
        elif self is other:
            return True
        else:
            return self.__dict__ == other.__dict__

    def __str__(self):
        return ('Hash: {}'.format(self.hash) + '\n'
                'Author: {}'.format(self.author.name) + '\n'
                'Author email: {}'.format(self.author.email) + '\n'
                'Committer: {}'.format(self.committer.name) + '\n'
                'Committer email: {}'.format(self.committer.email) + '\n'
                'Author date: {}'.format(self.author_date.strftime("%Y-%m-%d %H:%M:%S")) + '\n'
                'Committer date: {}'.format(self.committer_date.strftime("%Y-%m-%d %H:%M:%S")) + '\n'
                'Message: {}'.format(self.msg) + '\n'
                'Parent: {}'.format("\n".join(map(str, self.parents))) + '\n'
                'Merge: {}'.format(self.merge) + '\n'
                'Modifications: \n{}'.format("\n".join(map(str, self.modifications))) + '\n'
                'Branches: \n{}'.format("\n".join(map(str, self.branches))) + '\n'
                'In main branch: {}'.format(self.in_main_branch)
                )


class ChangeSet:
    def __init__(self, id: str, date: datetime):
        """
        Light-weight version of the commit, storing only the hash and the date. Used for filter out
        commits before asking for more complex information (like diff and source code).

        :param str id: hash of the commit
        :param date: date of the commit
        """
        self.id = id
        self.date = date

    def __eq__(self, other):
        if not isinstance(other, ChangeSet):
            return NotImplemented
        elif self is other:
            return True
        else:
            return self.__dict__ == other.__dict__


# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
import os


class ModificationType(Enum):
    ADD = 1,
    COPY = 2,
    RENAME = 3,
    DELETE = 4,
    MODIFY = 5


class Modification:
    def __init__(self, old_path: str, new_path: str, change_type: ModificationType, diff: str, source_code: str):
        """
        Initialize a modification. A modification carries on information regarding
        the changed file.

        :param old_path: old path of the file (can be null if the file is added)
        :param new_path: new path of the file (can be null if the file is deleted)
        :param change_type: type of the change
        :param diff: diff of the change
        :param source_code: source code of the file (can be null if the file is deleted)
        """

        self.old_path = old_path
        self.new_path = new_path
        self.change_type = change_type
        self.diff = diff
        self.source_code = source_code
        self.added = 0
        self.removed = 0
        self.filename = self._get_filename()

        for line in diff.replace('\r', '').split("\n"):
            if line.startswith('+') and not line.startswith('+++'):
                self.added += 1
            if line.startswith('-') and not line.startswith('---'):
                self.removed += 1

    def _get_filename(self) -> str:
        if self.new_path is not None and self.new_path != "/dev/null":
            path = self.new_path
        else:
            path = self.old_path

        if os.sep not in path:
            return path

        filename = path.split(os.sep)
        return filename[-1]

    def __eq__(self, other):
        if not isinstance(other, Modification):
            return NotImplemented
        elif self is other:
            return True
        else:
            return self.__dict__ == other.__dict__

    def __str__(self):
        return (
            'MODIFICATION\n' +
            'Old Path: {}\n'.format(self.old_path) +
            'New Path: {}\n'.format(self.new_path) +
            'Type: {}\n'.format(self.change_type.name) +
            'Diff: {}\n'.format(self.diff) +
            'Source code: {}\n'.format(self.source_code)
        )


# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s', level=logging.INFO)

from pydriller.domain.modification import Modification, ModificationType


def test_should_get_adds_and_removes():

    diff = "+++ diff info\n" \
       + "+ new line\n" \
       + "- removed line\n" \
       + "+ new line\n" \
       + "\n" \
       + "+ new line\n" \
       + "- removed line\n" \
       + "+++ more git stuff\n" \
       + "--- more git stuff\n"

    m = Modification("old", "new", ModificationType.ADD, diff, "class Java {} ");

    assert 3 == m.added
    assert 2 == m.removed


def test_should_get_name_of_file_even_when_deleted():
    m1 = Modification("/a/b/Class.java", "/dev/null", ModificationType.DELETE, "bla bla", "bla bla");
    assert "Class.java" == m1.filename

    m2 = Modification("/a/b/Class.java", "/a/b/Class.java", ModificationType.MODIFY, "bla bla", "bla bla");
    assert "Class.java" == m2.filename

    m3 = Modification("/dev/null", "/a/b/Class.java", ModificationType.ADD, "bla bla", "bla bla");
    assert "Class.java" == m3.filename



# Copyright 2018 Davide Spadini
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
from typing import List, Dict, Tuple
from git import Git, Repo, Diff, GitCommandError
from pydriller.domain.commit import Commit, ChangeSet
from pydriller.domain.developer import Developer
from pydriller.domain.modification import ModificationType
from threading import Lock

logger = logging.getLogger(__name__)


class GitRepository:
    def __init__(self, path: str):
        """
        Init the Git Repository.

        :param path: path to the repository
        """
        self.path = path
        self.main_branch = None
        self.lock = Lock()

    def _open_git(self) -> Git:
        self._open_repository()
        return Git(self.path)

    def _open_repository(self) -> Repo:
        repo = Repo(self.path)
        if self.main_branch is None:
            self._discover_main_branch(repo)
        return repo

    def _discover_main_branch(self, repo):
        self.main_branch = repo.active_branch.name

    def get_head(self) -> ChangeSet:
        """
        Get the head commit.

        :return: ChangeSet of the head commit
        """
        repo = self._open_repository()
        head_commit = repo.head.commit
        return ChangeSet(head_commit.hexsha, head_commit.committed_datetime)

    def get_change_sets(self) -> List[ChangeSet]:
        """
        Return the list of all the commits in the repo.

        :return: List[ChangeSet], the list of all the commits in the repo
        """
        return self._get_all_commits()

    def _get_all_commits(self) -> List[ChangeSet]:
        repo = self._open_repository()
        commit_list = list(repo.iter_commits())

        change_sets = []
        for commit in commit_list:
            committer_date = commit.committed_datetime
            change_sets.append(ChangeSet(commit.hexsha, committer_date))
        return change_sets

    def get_commit(self, commit_id: str) -> Commit:
        """
        Get the specified commit.

        :param commit_id: hash of the commit to analyze
        :return: Commit
        """
        git = self._open_git()
        repo = self._open_repository()
        commit = repo.commit(commit_id)

        author = Developer(commit.author.name, commit.author.email)
        committer = Developer(commit.committer.name, commit.committer.email)
        author_timezone = commit.author_tz_offset
        committer_timezone = commit.committer_tz_offset

        msg = commit.message.strip()
        commit_hash = commit.hexsha

        author_date = commit.authored_datetime
        committer_date = commit.committed_datetime

        merge = True if len(commit.parents) > 1 else False

        parents = []
        for p in commit.parents:
            parents.append(p.hexsha)

        branches = self._get_branches(git, commit_hash)
        is_in_main_branch = self.main_branch in branches

        the_commit = Commit(commit_hash, author, committer, author_date, committer_date, author_timezone,
                            committer_timezone, msg,
                            parents, merge, branches, is_in_main_branch)

        if len(parents) > 0:
            parent = repo.commit(parents[0])
            diff_index = parent.diff(commit, create_patch=True)
        else:
            parent = repo.tree('4b825dc642cb6eb9a060e54bf8d69288fbee4904')
            diff_index = parent.diff(commit.tree, create_patch=True)

        self._parse_diff(diff_index, the_commit)

        return the_commit

    def _parse_diff(self, diff_index, the_commit) -> None:
        for d in diff_index:
            old_path = d.a_path
            new_path = d.b_path
            change_type = self._from_change_to_modification_type(d)
            sc = ''
            diff_text = ''
            try:
                sc = d.b_blob.data_stream.read().decode('utf-8')
                diff_text = d.diff.decode('utf-8')
            except (UnicodeDecodeError, AttributeError, ValueError):
                logger.debug('Could not load source code or the diff of a file in commit {}'.format(the_commit.hash))

            the_commit.add_modifications(old_path, new_path, change_type, diff_text, sc)

    def _get_branches(self, git: Git, commit_hash: str) -> set():
        branches = set()
        for branch in set(git.branch('--contains', commit_hash).split('\n')):
            branches.add(branch.strip().replace('* ', ''))
        return branches

    def _from_change_to_modification_type(self, d: Diff):
        if d.new_file:
            return ModificationType.ADD
        elif d.deleted_file:
            return ModificationType.DELETE
        elif d.renamed_file:
            return ModificationType.RENAME
        elif d.a_blob and d.b_blob and d.a_blob != d.b_blob:
            return ModificationType.MODIFY

    def checkout(self, _hash: str) -> None:
        """
        Checkout the repo at the speficied commit.
        BE CAREFUL: this will change the state of the repo, hence it should *not*
        be used with more than 1 thread.

        :param _hash: commit hash to checkout
        """
        with self.lock:
            git = self._open_git()
            self._delete_tmp_branch()
            git.checkout('-f', _hash, b='_PD')

    def _delete_tmp_branch(self) -> None:
        repo = self._open_repository()
        try:
            repo.delete_head('_PD')
        except GitCommandError:
            logger.debug("Branch _PD not found")

    def files(self) -> List[str]:
        """
        Obtain the list of the files (excluding .git directory).

        :return: List[str], the list of the files
        """
        _all = []
        for path, subdirs, files in os.walk(self.path):
            if '.git' in path:
                continue
            for name in files:
                _all.append(os.path.join(path, name))
        return _all

    def reset(self) -> None:
        """
        Reset the state of the repo, checking out the main branch and discarding
        local changes (-f option).

        """
        with self.lock:
            git = self._open_git()
            git.checkout('-f', self.main_branch)
            self._delete_tmp_branch()

    def total_commits(self) -> int:
        """
        Calculate total number of commits.

        :return: the total number of commits
        """
        return len(self.get_change_sets())

    def get_commit_from_tag(self, tag: str) -> Commit:
        """
        Obtain the tagged commit.

        :param str tag: the tag
        :return: Commit commit: the commit the tag referred to
        """
        repo = self._open_repository()
        try:
            selected_tag = repo.tags[tag]
            return self.get_commit(selected_tag.commit.hexsha)
        except (IndexError, AttributeError):
            logger.debug('Tag {} not found'.format(tag))
            raise

    def parse_diff(self, diff: str) -> Dict[str, List[Tuple[int, str]]]:
        """
        Given a diff, returns a dictionary with the added and deleted lines.
        The dictionary has 2 keys: "added" and "deleted", each containing the
        corresponding added or deleted lines. For both keys, the value is a list
        of Tuple (int, str), corresponding to (number of line in the file, actual line).


        :param str diff: diff of the commit
        :return: Dictionary
        """
        lines = diff.split('\n')
        modified_lines = {'added': [], 'deleted': []}

        count_deletions = 0
        count_additions = 0

        for line in lines:
            line = line.rstrip()
            count_deletions += 1
            count_additions += 1

            if line.startswith('@@'):
                count_deletions, count_additions = self._get_line_numbers(line)

            if line.startswith('-'):
                modified_lines['deleted'].append((count_deletions, line[1:]))
                count_additions -= 1

            if line.startswith('+'):
                modified_lines['added'].append((count_additions, line[1:]))
                count_deletions -= 1

        return modified_lines

    def _get_line_numbers(self, line):
        token = line.split(" ")
        numbers_old_file = token[1]
        numbers_new_file = token[2]
        delete_line_number = int(numbers_old_file.split(",")[0].replace("-", "")) - 1
        additions_line_number = int(numbers_new_file.split(",")[0]) - 1
        return delete_line_number, additions_line_number
